from tax_module.models import PricingType, Price, TaxableItem


def create_taxable_item(price: Price = None, region: str = None, item_type: str = None):
    class FakeTaxableItem(TaxableItem):
        @property
        def price(self):
            return price or create_price()

        @property
        def region(self):
            return region or "NCR"

        @property
        def item_type(self):
            return item_type or "SERVICE"

    return FakeTaxableItem()


def create_price(amount: int = None, pricing_type: PricingType = None):
    return Price(
        amount=amount or 100, pricing_type=pricing_type or PricingType.TAX_INCLUSIVE
    )



from typing import Dict, List

from tax_module.models import Tax

CGST_PERCENTAGE_FOR_SERVICE = 6
SGST_PERCENTAGE_FOR_SERVICE = 6
IGST_PERCENTAGE_FOR_SERVICE = 12

INDIAN_TAX_RULE_BOOK: Dict[str, Dict[str, List[Tax]]] = {
    "MAHARASHTRA": {
        "SERVICE": [
            Tax(tax_type="CGST", tax_percentage=CGST_PERCENTAGE_FOR_SERVICE),
            Tax(tax_type="SGST", tax_percentage=SGST_PERCENTAGE_FOR_SERVICE),
        ]
    },
    "NCR": {
        "SERVICE": [Tax(tax_type="IGST", tax_percentage=IGST_PERCENTAGE_FOR_SERVICE)]
    },
    "TAMIL NADU": {
        "SERVICE": [
            Tax(tax_type="CGST", tax_percentage=CGST_PERCENTAGE_FOR_SERVICE),
            Tax(tax_type="SGST", tax_percentage=SGST_PERCENTAGE_FOR_SERVICE),
        ]
    },
    "TELANGANA": {
        "SERVICE": [
            Tax(tax_type="CGST", tax_percentage=CGST_PERCENTAGE_FOR_SERVICE),
            Tax(tax_type="SGST", tax_percentage=SGST_PERCENTAGE_FOR_SERVICE),
        ]
    },
    "WEST BENGAL": {
        "SERVICE": [
            Tax(tax_type="CGST", tax_percentage=CGST_PERCENTAGE_FOR_SERVICE),
            Tax(tax_type="SGST", tax_percentage=SGST_PERCENTAGE_FOR_SERVICE),
        ]
    },
}



import pytest

from tax_module.base import TaxModule
from tax_module.models import PricingType, Tax
from tax_module.errors import NoTaxRuleFoundError
from tests.factory import create_taxable_item, create_price


def test_tax_module_gives_single_tax_for_taxable_item():
    fake_rule_book = {
        "MY_REGION": {"TAXABLE_ITEM_TYPE": [Tax(tax_percentage=10, tax_type="MY_TAX")]}
    }

    tax_module = TaxModule(rule_book=fake_rule_book)
    taxable_item = create_taxable_item(
        region="MY_REGION", item_type="TAXABLE_ITEM_TYPE"
    )
    taxes = tax_module.get_taxes_for_taxable_item(taxable_item)

    assert len(taxes) == 1
    assert taxes[0].tax.tax_percentage == 10
    assert taxes[0].tax.tax_type == "MY_TAX"
    assert taxes[0].amount == 9.09


def test_tax_module_gives_multiple_taxes_for_taxable_item():
    fake_rule_book = {
        "MY_REGION": {
            "TAXABLE_ITEM_TYPE": [
                Tax(tax_percentage=10, tax_type="MY_TAX"),
                Tax(tax_percentage=11, tax_type="MY_TAX1"),
            ]
        }
    }

    tax_module = TaxModule(rule_book=fake_rule_book)
    taxable_item = create_taxable_item(
        region="MY_REGION", item_type="TAXABLE_ITEM_TYPE"
    )
    taxes = tax_module.get_taxes_for_taxable_item(taxable_item)

    assert len(taxes) == 2
    assert taxes[0].tax.tax_percentage == 10
    assert taxes[0].tax.tax_type == "MY_TAX"
    assert taxes[0].amount == 8.26
    assert taxes[1].tax.tax_percentage == 11
    assert taxes[1].tax.tax_type == "MY_TAX1"
    assert taxes[1].amount == 9.09


def test_tax_module_raises_exception_incase_no_rule_exists():
    fake_rule_book = {
        "MY_REGION": {
            "OTHER_TAXABLE_ITEM_TYPE": [Tax(tax_percentage=10, tax_type="MY_TAX")]
        }
    }

    tax_module = TaxModule(rule_book=fake_rule_book)
    taxable_item = create_taxable_item(
        region="MY_REGION", item_type="TAXABLE_ITEM_TYPE"
    )
    with pytest.raises(NoTaxRuleFoundError):
        tax_module.get_taxes_for_taxable_item(taxable_item)


def test_tax_module_for_price_exclusive_taxable_item():
    fake_rule_book = {
        "MY_REGION": {"TAXABLE_ITEM_TYPE": [Tax(tax_percentage=10, tax_type="MY_TAX")]}
    }

    tax_module = TaxModule(rule_book=fake_rule_book)
    taxable_item = create_taxable_item(
        region="MY_REGION",
        item_type="TAXABLE_ITEM_TYPE",
        price=create_price(amount=100, pricing_type=PricingType.TAX_EXCLUSIVE),
    )
    taxes = tax_module.get_taxes_for_taxable_item(taxable_item)

    assert len(taxes) == 1
    assert taxes[0].tax.tax_percentage == 10
    assert taxes[0].tax.tax_type == "MY_TAX"
    assert taxes[0].amount == 10.0



from typing import List

from tax_module.models import AppliedTax, TaxableItem
from tax_module.errors import NoTaxRuleFoundError


class TaxModule:
    def __init__(self, rule_book: dict):
        self._rule_book = rule_book  # It is a dictionary which tells which tax_type applies to which zone

    def get_taxes_for_taxable_item(self, taxable_item: TaxableItem) -> List[AppliedTax]:
        try:
            taxes_to_apply = self._rule_book[taxable_item.region][
                taxable_item.item_type
            ]
        except KeyError:
            raise self._mk_exc(
                region=taxable_item.region, item_type=taxable_item.item_type
            )
        sum_of_taxes = sum(tax.tax_percentage for tax in taxes_to_apply)
        return [
            AppliedTax(tax=tax, price=taxable_item.price, sum_of_taxes=sum_of_taxes)
            for tax in taxes_to_apply
        ]

    def _mk_exc(self, region: str, item_type: str) -> NoTaxRuleFoundError:
        return NoTaxRuleFoundError(
            f"""
        No Tax Rule Found for
        Region: {region}
        ItemType:{item_type}
        """
        )


def get_tax_module(rule_book):
    return TaxModule(rule_book=rule_book)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import List


class PricingType(Enum):
    TAX_INCLUSIVE = "TAX_INCLUSIVE"
    TAX_EXCLUSIVE = "TAX_EXCLUSIVE"


@dataclass
class Price:
    amount: int
    pricing_type: PricingType

    @property
    def is_tax_inclusive(self):
        return self.pricing_type == PricingType.TAX_INCLUSIVE


@dataclass
class Tax:
    tax_percentage: int
    tax_type: str


class AppliedTax:
    def __init__(self, tax: Tax, price: Price, sum_of_taxes: int):
        tax_exclusive_amount = (
            (price.amount * 100 / (100 + sum_of_taxes))
            if price.is_tax_inclusive
            else price.amount
        )
        self.tax = tax
        self.amount = round((tax_exclusive_amount * self.tax.tax_percentage) / 100, 2)


class TaxableItem(ABC):
    @property
    @abstractmethod
    def price(self) -> Price:
        pass

    @property
    @abstractmethod
    def region(self) -> str:
        pass

    @property
    @abstractmethod
    def item_type(self) -> str:
        pass


@dataclass
class TaxedItem(ABC):
    taxes: List[AppliedTax]


class TaxableItemType(Enum):
    GOOD = "GOOD"
    SERVICE = "SERVICE"